import os
import signal
import sys
from collections import namedtuple

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from .materials.delay_data import DelayData
from .materials.energy_data import EnergyData
from .materials.fit_data import FitData
from .materials.image_data import ImageData
from .materials.metrics_data import MetricsData
from .materials.position_data import PositionData
from .materials.reflectivity_data import ReflectivityData
from .materials.transmission_data import TransmissionData
from .materials.roi_data import ROIData
from .services.export_service import ExportService
from .services.fit_service import FitService
from .services.io_service import IOService
from .services.metrics_service import MetricsService
from .services.timing_service import TimingService
from .services.profiler_service import ProfilerService
from .tools.ramanometer_tool import RamanometerTool

__application_name__ = "Ramanometer"
__version__ = "0.3.9"
__elogs__ = {"FLASH1 Raman": "fl1_ramlog"}
__export__directory__ = os.path.expanduser("~") + "/Ramanometer/Data/"


def startup_application():
    signal.signal(signal.SIGINT, signal.SIG_DFL)  # handles Ctrl-C interrupts via default signal (core dump and exit)

    resources_path = os.path.dirname(os.path.abspath(__file__)) + "/resources/"

    application = QApplication([__application_name__])
    application.setWindowIcon(QIcon(resources_path + "ramanometer.png"))

    image_data = ImageData(address="TTF2.FEL/HASVUVPGFW3.CAM/PIXIS/IMAGE_EXT_ZMQ")
    energy_data = EnergyData(readback_address="TTF2.FEL/HASY.MONOCHR/MONOCHR/ENERGY",
                             setpoint_address="TTF2.FEL/HASY.MONOCHR/MONOCHR/ENERGYTOSET",
                             setpoint_change_address="TTF2.FEL/HASY.MONOCHR/MONOCHR/CHANGEENERGY")
    fit_data = FitData()
    metrics_data = MetricsData()
    roi_data = ROIData(file=resources_path + "roi.ini")
    delay_data = DelayData(readback_address="FLASH.SYNC/LASER.LOCK.EXP/F1.PG.OSC/FMC0.MD22.1.POSITION.RD",
                           setpoint_address="FLASH.SYNC/LASER.LOCK.EXP/F1.PG.OSC/FMC0.MD22.1.POSITION_SET.WR",
                           encoder_address="FLASH.SYNC/LASER.LOCK.EXP/F1.PG.OSC/FMC0.MD22.1.ENCODER_POSITION.RD")
    reflectivity_data = ReflectivityData(adc_address="FLASH.FEL/ADC.SIS.PG/EXP1.CH06/CH00.TD",
                                         bunch_number_address="FLASH.DIAG/TOROID.ML/3GUN/NUMBEROFBUNCHES.FLASH1")
    alignment_position_data = PositionData(readback_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR7.MOT1/FPOS",
                                           setpoint_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR7.MOT1/FPOS.SET",
                                           setpoint_change_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR7.MOT1/CMD",
                                           reference_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR7.MOT1/POS.SET")
    profiler_positions = namedtuple("ProfilerPositionsData", ("horizontal", "vertical"))(
        PositionData(readback_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR8.MOT1/FPOS",
                     setpoint_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR8.MOT1/FPOS.SET",
                     setpoint_change_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR8.MOT1/CMD",
                     reference_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR8.MOT1/POS.SET"),
        PositionData(readback_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR6.MOT1/FPOS",
                     setpoint_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR6.MOT1/FPOS.SET",
                     setpoint_change_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR6.MOT1/CMD",
                     reference_address="FLASH.FEL/PG1.TRIXS.MOTOR/MOTOR6.MOT1/POS.SET"))
    transmission_data = TransmissionData(adc_address="FLASH.FEL/ADC.SIS.PG/EXP1.CH06/CH00.TD")

    io_service = IOService(image_data, energy_data, alignment_position_data)
    export_service = ExportService(directory=__export__directory__)
    fit_service = FitService(image_data, fit_data)
    metrics_service = MetricsService(metrics_data, fit_data, energy_data, alignment_position_data)
    timing_service = TimingService(io_service, delay_data, reflectivity_data)
    profiler_service = ProfilerService(io_service, profiler_positions, transmission_data)

    ramanometer_tool = RamanometerTool(__application_name__, __version__, __elogs__,
                                       io_service, fit_service, metrics_service, timing_service, profiler_service,
                                       export_service, roi_data)
    ramanometer_tool.show_window()

    sys.exit(application.exec())
