import time
from collections import namedtuple

import numpy as np


class PositionData:
    _Position = namedtuple("Position", ("values", "errors"))

    def __init__(self, readback_address, setpoint_address, setpoint_change_address, reference_address):
        self._readback_address = readback_address
        self._setpoint_address = setpoint_address
        self._setpoint_change_address = setpoint_change_address
        self._reference_address = reference_address
        self._readback_position = None
        self._is_position_to_set = False
        self._is_ready_for_new_data = None
        self._is_ready_for_new_scan_point = None
        self._position = None
        self._scan_range = None
        self._samples_per_step = None
        self._tolerance = None
        self._timeout = None
        self._reference = None
        self._is_ready_to_check_position = None
        self._scan_step = None
        self._last_time_position_to_set = None
        self._positions = []

    @property
    def readback_address(self):
        return self._readback_address

    @property
    def setpoint_address(self):
        return self._setpoint_address

    @property
    def setpoint_change_address(self):
        return self._setpoint_change_address

    @property
    def reference_address(self):
        return self._reference_address

    @property
    def readback_position(self):
        return self._readback_position

    @property
    def is_position_to_set(self):
        return self._is_position_to_set

    @property
    def is_ready_for_new_data(self):
        return self._is_ready_for_new_data

    @property
    def is_ready_for_new_scan_point(self):
        return self._is_ready_for_new_scan_point

    @property
    def reference(self):
        return self._reference

    @property
    def scan_step(self):
        return self._scan_step

    @property
    def is_ready_to_check_position(self):
        return self._is_ready_to_check_position

    @property
    def position(self):
        return self._position

    def initialize(self, scan_range, samples_per_step, tolerance, timeout):
        self._scan_range = scan_range
        self._samples_per_step = samples_per_step
        self._tolerance = tolerance
        self._timeout = timeout
        self._is_position_to_set = True
        self._is_ready_for_new_scan_point = True
        self._scan_step = 0
        self._positions.clear()
        self._position = None

    def reset(self):
        self._is_position_to_set = False
        self._is_ready_for_new_scan_point = False
        self._is_ready_for_new_data = False

    def is_ready_for_new_setpoint(self):
        if self._is_ready_for_new_scan_point:
            self._is_ready_to_check_position = False
            self._is_ready_for_new_scan_point = False
            self._is_ready_for_new_data = False
            return True
        return False

    def get_next_setpoint(self):
        next_setpoint = self._scan_range[self._scan_step]
        self._scan_step += 1
        return next_setpoint

    def set_reference(self, reference):
        self._reference = reference

    def set_ready_to_check(self):
        self._is_ready_to_check_position = True
        self._last_time_position_to_set = time.time()

    def check_position(self, readback_position, setpoint_position):
        delta_position = abs(readback_position - setpoint_position)
        delta_time = time.time() - self._last_time_position_to_set
        if delta_position <= self._tolerance or delta_time >= self._timeout:
            self._is_ready_for_new_data = True

    def update(self, position_data):
        self._readback_position = position_data
        if self._is_ready_for_new_data:
            self._positions.append(position_data)
            if len(self._positions) % self._samples_per_step == 0:
                self._is_ready_for_new_scan_point = True
                position = np.array(np.split(np.array(self._positions), self._scan_step))
                position_values_and_errors = np.stack((np.nanmean(position, axis=1), np.nanstd(position, axis=1)),
                                                      axis=-1)
                self._position = self._Position(position_values_and_errors[:, 0], position_values_and_errors[:, 1])

    def get_scanning_progress(self):
        return len(self._positions) / (self._samples_per_step * len(self._scan_range)) * 100

    def is_not_empty(self):
        return self._position is not None
