import numpy as np
from doocspie.util import Colors
from pyqtgraph import GraphicsLayoutWidget, ErrorBarItem, PlotDataItem, LegendItem, QtCore


class TimingPlotter:
    _DESY_ORANGE = np.array(Colors.DESY.ORANGE) * 255
    _DESY_CYAN = np.array(Colors.DESY.CYAN) * 255
    _DESY_GREY = np.array(Colors.DESY.GRAY_1) * 255

    def __init__(self):
        self._widget = GraphicsLayoutWidget()

        self._timing_plot = self._widget.addPlot()
        self._timing_error = ErrorBarItem(x=[], y=[])
        self._timing_line = PlotDataItem()
        self._timing_fit = PlotDataItem()

        self._timing_plot.addItem(self._timing_error)
        self._timing_plot.addItem(self._timing_line)
        self._timing_plot.addItem(self._timing_fit)

        self._legend = LegendItem(offset=(70, 400))
        self._current_legend = None

        self._initialize_plots()

    def _initialize_plots(self):
        self._timing_plot.setMenuEnabled(False)
        self._timing_plot.showAxis("top", show=True)
        self._timing_plot.showAxis("right", show=True)
        for pos in "top", "right", "left", "bottom":
            self._timing_plot.getAxis(pos).setStyle(showValues=False, tickLength=0)
        self._timing_error.setData(pen={"color": self._DESY_GREY, "width": 2})
        self._timing_line.setData(symbol="o", pen=None,
                                  symbolBrush=tuple(self._DESY_ORANGE), symbolPen=tuple(self._DESY_ORANGE))
        self._timing_fit.setData(pen={"color": self._DESY_CYAN, "width": 2})

    @property
    def widget(self):
        return self._widget

    def set_timing(self, delay, reflectivity):
        self._timing_error.setData(x=delay.values, y=reflectivity.values,
                                   width=2 * delay.errors, height=2 * reflectivity.errors)
        self._timing_line.setData(x=delay.values, y=reflectivity.values)

        self._timing_plot.getAxis("left").setStyle(showValues=True)
        self._timing_plot.setLabel("left", "Reflectivity (%)")
        self._timing_plot.getAxis("bottom").setStyle(showValues=True)
        self._timing_plot.setLabel("bottom", "Encoder Delay (ps)")

    def set_fit(self, fit):
        self._timing_fit.setData(x=fit.delay, y=fit.reflectivity)
        if self._current_legend != fit.legend:
            self._legend.removeItem(self._current_legend)
            self._current_legend = fit.legend
            self._legend.setParentItem(self._timing_plot)
            self._legend.addItem(self._timing_fit, fit.legend)
            self._legend.setGeometry(QtCore.QRectF(0, 0, self._legend.width() + 10, self._legend.height()))

    def clear_fit(self):
        self._timing_fit.clear()
        self._legend.removeItem(self._current_legend)
        self._current_legend = None
        if self._legend.scene():
            self._legend.scene().removeItem(self._legend)

    def clear(self):
        self._timing_error.setData(x=[], y=[], width=np.array([]), height=np.array([]))
        self._timing_line.clear()
        self._timing_fit.clear()
        self._timing_plot.setLabel("left", "")
        self._timing_plot.getAxis("left").setStyle(showValues=False, tickLength=0)
        self._timing_plot.setLabel("bottom", "")
        self._timing_plot.getAxis("bottom").setStyle(showValues=False, tickLength=0)
