import warnings

import numpy as np
from scipy.optimize import curve_fit, OptimizeWarning
from scipy.special import erf


class FitService:
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    warnings.filterwarnings("ignore", category=OptimizeWarning)

    def __init__(self, image_data, fit_data):
        self._image_data = image_data
        self._fit_data = fit_data

    @property
    def fit_data(self):
        return self._fit_data

    def update(self):
        self._fit_data.update(self._get_gaussian_fit, self._image_data.projections_values.horizontal,
                              self._image_data.projections_values.vertical)

    def _get_gaussian_fit(self, projection):
        fit_parameters = self._get_gaussian_fit_parameters(projection.x, projection.y)
        if fit_parameters is not None:
            x = np.linspace(projection.x[0], projection.x[-1], num=int(10 * len(projection.x)))
            fit = x, self._gauss_model(x, *fit_parameters)
            return fit, fit_parameters
        return None, None

    def _get_gaussian_fit_parameters(self, x, y):
        try:
            offset = y[0]
            slope = (y[-1] - y[0]) / (x[-1] - x[0])
            y_corrected = y - offset - slope * (x - x[0])  # correct offset and slope before guessing start parameters
            amplitude = np.nanmax(y_corrected)
            position = x[np.nanargmax(y_corrected)]
            sigma = np.abs((x[-1] - x[0]) / 50)  # sigma guess is found empirically and should be used with caution
            guess = (offset, slope, amplitude, position, sigma)
            valid = ~(np.isnan(x) | np.isnan(y))
            fit_parameters, *_ = curve_fit(self._gauss_model, x[valid], y[valid], p0=guess)
            fit_parameters[4] = np.abs(fit_parameters[4])
            return fit_parameters
        except (RuntimeError, ValueError, TypeError, IndexError):
            return None

    @staticmethod
    def _gauss_model(x, offset, slope, amplitude, mean, sigma):
        return offset + slope * x + amplitude * np.exp(-(x - mean) ** 2 / 2 / sigma ** 2)

    def get_fit_for(self, fit_model, x, y, y_error):
        fit_parameters, fit_covariance = self._get_fit_parameters_for(fit_model, x, y, y_error)
        if fit_parameters is not None and fit_covariance is not None:
            x = np.linspace(x[0], x[-1], num=int(10 * len(x)))
            fit = x, fit_model(x, *fit_parameters)
            legend = self._get_legend_for(fit_model, fit_parameters, fit_covariance)
            return fit, legend
        return None, None

    @staticmethod
    def _get_fit_parameters_for(fit_model, x, y, y_error):
        try:
            y_error = y_error.copy()
            y_error[y_error == 0] = 1e-15  # curve fit cannot handle zero error
            valid = ~(np.isnan(x) | np.isnan(y) | np.isnan(y_error))
            fit_parameters, fit_covariance, *_ = curve_fit(fit_model, x[valid], y[valid], sigma=y_error[valid])
            return fit_parameters, fit_covariance
        except (RuntimeError, ValueError, TypeError, IndexError):
            return None, None

    def get_fit_model(self, model):
        if model == "Constant":
            return self._constant_model
        if model == "Linear":
            return self._linear_model
        if model == "Quadratic":
            return self._quadratic_model
        return None

    def _get_legend_for(self, fit_model, fit_parameters, fit_covariance):
        legend = None
        if fit_model.__name__ == self._constant_model.__name__:
            legend = f"average: {fit_parameters[0]:.2f} px"
        if fit_model.__name__ == self._linear_model.__name__:
            inverse_slope = 1 / fit_parameters[1]
            inverse_slope_error = np.sqrt(np.diag(fit_covariance))[1] * inverse_slope ** 2
            intercept = -fit_parameters[0] / fit_parameters[1]
            intercept_error = abs(intercept) * np.sqrt(
                (np.sqrt(np.diag(fit_covariance))[0] / fit_parameters[0]) ** 2 +
                (np.sqrt(np.diag(fit_covariance))[1] / fit_parameters[1]) ** 2)
            legend = fr"inverse slope: {1000 * inverse_slope:.1f} &#177; {1000 * inverse_slope_error:.1f} meV/px, "
            legend += fr"intercept: {intercept:.1f} &#177; {intercept_error:.1f} eV"
        if fit_model.__name__ == self._quadratic_model.__name__:
            extremum_position = -fit_parameters[1] / 2 / fit_parameters[2]
            extremum_value = self._quadratic_model(extremum_position, *fit_parameters)
            legend = f"extremum: ({extremum_position:.2f} eV, {extremum_value:.2f} px)"
        return legend

    @classmethod
    def _constant_model(cls, x, offset):
        return cls._linear_model(x, offset, 0)

    @classmethod
    def _linear_model(cls, x, offset, slope):
        return cls._quadratic_model(x, offset, slope, 0)

    @classmethod
    def _quadratic_model(cls, x, offset, slope, curvature):
        return offset + slope * x + curvature * x ** 2

    def get_timing_fit(self, delay, reflectivity):
        fit_parameters = self._get_fit_of_decay_model_for(delay.values, reflectivity.values)
        if fit_parameters is not None:
            x = np.linspace(delay.values[0], delay.values[-1], num=int(10 * len(delay.values)))
            fit = x, self._decay_model(x, *fit_parameters)
            time_zero = fit_parameters[0]
            resolution = abs(fit_parameters[4]) * 1e3
            legend = f"time zero at {time_zero:.2f} ps, r.m.s resolution of {resolution:.0f} fs"
            return fit, legend
        return None, None

    def _get_fit_of_decay_model_for(self, x, y):
        try:
            delay_offset = np.nanmin(x) + (np.nanmax(x) - np.nanmin(x)) / 2
            offset = np.nanmax(y)
            amplitude = (np.nanmax(y) - np.nanmin(y)) / 2
            tau = 10
            sigma = 0.1
            guess = [delay_offset, offset, amplitude, tau, sigma]
            valid = ~(np.isnan(x) | np.isnan(y))
            fit_parameters, *_ = curve_fit(self._decay_model, x[valid], y[valid], p0=guess)
            return fit_parameters
        except (RuntimeError, ValueError, TypeError, IndexError):
            return None

    @staticmethod
    def _decay_model(delay, delay_offset, offset, amplitude, tau, sigma):
        return (offset - amplitude * np.exp(-((delay - delay_offset) - sigma ** 2 / 2 / tau) / tau) *
                (1 + erf(((delay - delay_offset) - sigma ** 2 / tau) / np.sqrt(2) / abs(sigma))))

    def get_alignment_fit(self, x, y, y_error):
        fit_model = self.get_fit_model("Quadratic")
        fit_parameters, fit_covariance = self._get_fit_parameters_for(fit_model, x, y, y_error)
        if fit_parameters is not None and fit_covariance is not None:
            x = np.linspace(x[0], x[-1], num=int(10 * len(x)))
            fit = x, fit_model(x, *fit_parameters)
            extremum_position = -fit_parameters[1] / 2 / fit_parameters[2]
            extremum_value = self._quadratic_model(extremum_position, *fit_parameters)
            legend = f"extremum: ({extremum_position:.2f} mm, {extremum_value:.2f} px)"
            return fit, legend
        return None, None

    def get_profiler_fit(self, x, transmission):
        fit_parameters, fit_covariance = self._get_fit_of_step_model_for(x.values, transmission.values,
                                                                         transmission.errors)
        if fit_parameters is not None and fit_covariance is not None:
            x = np.linspace(x.values[0], x.values[-1], num=int(10 * len(x.values)))
            fit = x, self._step_model(x, *fit_parameters)
            mean = fit_parameters[2]
            mean_error = np.sqrt(np.diag(fit_covariance))[2]
            sigma = fit_parameters[3]
            sigma_error = np.sqrt(np.diag(fit_covariance))[3]
            legend = fr"position = {mean:.2f} &#177; {mean_error :.2f} mm, "
            legend += fr"r.m.s. width = {sigma:.2f} &#177; {sigma_error :.2f} mm"
            return fit, legend
        return None, None

    def _get_fit_of_step_model_for(self, x, y, y_error):
        try:
            offset = np.nanmin(y)
            amplitude = np.abs(y[-1] - y[0])
            mean = x[np.nanargmin(np.abs(y - offset - amplitude / 2))]
            sigma = x[np.nanargmin(np.abs(y - offset - amplitude / 2))]
            guess = [offset, amplitude, mean, sigma]
            valid = ~(np.isnan(x) | np.isnan(y) | np.isnan(y_error))
            error = y_error[valid]
            if np.all(error == 0):
                error = None
            fit_parameters, fit_covariance, *_ = curve_fit(self._step_model, x[valid], y[valid], sigma=error, p0=guess)
            return fit_parameters, fit_covariance
        except (RuntimeError, ValueError, TypeError, IndexError):
            return None, None

    @staticmethod
    def _step_model(x, offset, amplitude, mean, sigma):
        return offset + amplitude / 2 * (1 - erf((x - mean) / np.sqrt(2) / sigma))
