from PyQt5.QtCore import QRegExp, Qt
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import (QMainWindow, QVBoxLayout, QPushButton, QSizePolicy, QProgressBar, QRadioButton, QLineEdit,
                             QWidget, QHBoxLayout, QGroupBox, QLabel)

from ..plotter.profiler_plotter import ProfilerPlotter


class ProfilerToolUi(QMainWindow):
    START = "Start"

    def __init__(self):
        super().__init__()
        self._name = "Profiler"
        self._main_layout = QVBoxLayout()
        self._profiler_plotter = ProfilerPlotter()
        self._start_button = self._get_push_button_with(self.START)
        self._stop_button = self._get_push_button_with("Stop")
        self._progress_bar = self._get_progress_bar()
        self._horizontal_radio_button = QRadioButton("Horizontal", checked=True)
        self._vertical_radio_button = QRadioButton("Vertical")
        self._begin_line_edit = self._get_line_edit()
        self._end_line_edit = self._get_line_edit()
        self._number_of_steps_line_edit = self._get_line_edit_with("10", "[1-9][0-9]*")
        self._samples_per_step_line_edit = self._get_line_edit_with("10", "[1-9][0-9]*")
        self._tolerance_line_edit = self._get_line_edit_with("50", "[0-9]*")
        self._timeout_line_edit = self._get_line_edit_with("5.0", "[0-9]*[.][0-9]*")
        self._check_button = self._get_push_button_with("Check")
        self._create_layout()

    @staticmethod
    def _get_push_button_with(label):
        push_button = QPushButton(label)
        push_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        return push_button

    @staticmethod
    def _get_progress_bar():
        progress_bar = QProgressBar()
        progress_bar.setValue(0)
        return progress_bar

    @staticmethod
    def _get_line_edit():
        line_edit = QLineEdit()
        line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        line_edit.setAlignment(Qt.AlignRight)
        line_edit.setValidator(QRegExpValidator(QRegExp("[-]?[0-9]*[.][0-9]*")))
        line_edit.setMaxLength(5)
        line_edit.setMaximumWidth(45)
        return line_edit

    @staticmethod
    def _get_line_edit_with(default_value, regex):
        integer_line_edit = QLineEdit(default_value)
        integer_line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        integer_line_edit.setAlignment(Qt.AlignRight)
        integer_line_edit.setMaxLength(4)
        integer_line_edit.setMaximumWidth(40)
        integer_line_edit.setValidator(QRegExpValidator(QRegExp(regex)))
        return integer_line_edit

    def _create_layout(self):
        self._create_main_layout()
        self._create_plot_layout()
        self._create_controls_layout()
        self._create_progress_layout()

    def _create_main_layout(self):
        main_widget = QWidget()
        main_widget.setLayout(self._main_layout)
        self.setCentralWidget(main_widget)

    def _create_plot_layout(self):
        plot_layout = QHBoxLayout()
        plot_layout.addWidget(self._profiler_plotter.widget)
        self._main_layout.addLayout(plot_layout)

    def _create_controls_layout(self):
        controls_layout = QHBoxLayout()
        controls_layout.addWidget(self._create_control_group_box())
        controls_layout.addWidget(self._create_scan_dimension_group_box())
        controls_layout.addWidget(self._create_position_range_group_box())
        controls_layout.addWidget(self._create_profiler_setup_group_box())
        controls_layout.addWidget(self._create_pulse_settings_group_box())
        self._main_layout.addLayout(controls_layout)

    def _create_control_group_box(self):
        group_box = QGroupBox("Control")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._start_button)
        widgets_layout.addWidget(self._stop_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_scan_dimension_group_box(self):
        group_box = QGroupBox("Scan Dimension")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._horizontal_radio_button)
        widgets_layout.addWidget(self._vertical_radio_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_position_range_group_box(self):
        group_box = QGroupBox("Position Range")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Begin (mm):"))
        widgets_layout.addWidget(self._begin_line_edit)
        widgets_layout.addWidget(QLabel("End (mm):"))
        widgets_layout.addWidget(self._end_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_profiler_setup_group_box(self):
        group_box = QGroupBox("Profiler Setup")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Number of Steps:"))
        widgets_layout.addWidget(self._number_of_steps_line_edit)
        widgets_layout.addWidget(QLabel("Samples / Step:"))
        widgets_layout.addWidget(self._samples_per_step_line_edit)
        widgets_layout.addWidget(QLabel("Tolerance (\u00B5m):"))
        widgets_layout.addWidget(self._tolerance_line_edit)
        widgets_layout.addWidget(QLabel("Timeout (s):"))
        widgets_layout.addWidget(self._timeout_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_pulse_settings_group_box(self):
        group_box = QGroupBox("Pulse Settings")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._check_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_progress_layout(self):
        progress_layout = QHBoxLayout()
        progress_layout.addWidget(QLabel("Progress:"))
        progress_layout.addWidget(self._progress_bar)
        self._main_layout.addLayout(progress_layout)

    @property
    def name(self):
        return self._name

    @property
    def start_button(self):
        return self._start_button

    @property
    def stop_button(self):
        return self._stop_button

    @property
    def horizontal_radio_button(self):
        return self._horizontal_radio_button

    @property
    def vertical_radio_button(self):
        return self._vertical_radio_button

    @property
    def begin_line_edit(self):
        return self._begin_line_edit

    @property
    def end_line_edit(self):
        return self._end_line_edit

    @property
    def number_of_steps_line_edit(self):
        return self._number_of_steps_line_edit

    @property
    def samples_per_step_line_edit(self):
        return self._samples_per_step_line_edit

    @property
    def tolerance_line_edit(self):
        return self._tolerance_line_edit

    @property
    def timeout_line_edit(self):
        return self._timeout_line_edit

    @property
    def check_button(self):
        return self._check_button

    def set_progress(self, value):
        self._progress_bar.setValue(value)

    def plot_scan(self, position, transmission, plot_label):
        self._profiler_plotter.set_profile(position, transmission, plot_label)

    def clear(self):
        self._profiler_plotter.clear()
        self.clear_fit()

    def plot_fit(self, fit):
        self._profiler_plotter.set_fit(fit)

    def clear_fit(self):
        self._profiler_plotter.clear_fit()
