from doocspie.pyqt import ELogPrinter, Timer, messages

from .alignment_tool import AlignmentTool
from .logger_tool import LoggerTool
from .profiler_tool import ProfilerTool
from .profiler_pulse_settings_tool import ProfilerPulseSettingsTool
from .timing_pulse_settings_tool import TimingPulseSettingsTool
from .ramanometer_tool_ui import RamanometerToolUi
from .scanner_tool import ScannerTool
from .timing_tool import TimingTool
from .viewer_tool import ViewerTool


class RamanometerTool:

    def __init__(self, application_name, version, elogs, io_service, fit_service, metrics_service, timing_service,
                 profiler_service, export_service, roi_data):
        self._viewer_tool = ViewerTool(io_service.image_data, io_service.energy_data,
                                       io_service.alignment_position_data, fit_service, roi_data)
        self._logger_tool = LoggerTool(io_service.image_data, fit_service.fit_data, metrics_service.metrics_data)
        self._scanner_tool = ScannerTool(io_service.energy_data, metrics_service.metrics_data, fit_service)
        self._timing_tool = TimingTool(timing_service.delay_data, timing_service.reflectivity_data, fit_service)
        self._timing_pulse_settings_tool = TimingPulseSettingsTool(timing_service)
        self._alignment_tool = AlignmentTool(io_service.alignment_position_data, metrics_service.metrics_data,
                                             fit_service)
        self._profiler_tool = ProfilerTool(profiler_service, fit_service)
        self._profiler_pulse_settings_tool = ProfilerPulseSettingsTool(profiler_service)
        self._ramanometer_tool_ui = RamanometerToolUi(application_name, version, self._viewer_tool.ui,
                                                      self._logger_tool.ui, self._scanner_tool.ui, self._timing_tool.ui,
                                                      self._alignment_tool.ui, self._profiler_tool.ui)
        self._e_log_printer = ELogPrinter(self._ramanometer_tool_ui, printers=elogs, print_exceptions=True)
        self._main_concurrency_service = Timer(self._main_process_loop)
        self._timing_concurrency_service = Timer(self._timing_process_loop)
        self._profiler_concurrency_service = Timer(self._profiler_process_loop)
        self._io_service = io_service
        self._export_service = export_service(self._ramanometer_tool_ui)
        self._fit_service = fit_service
        self._metrics_service = metrics_service
        self._timing_service = timing_service
        self._profiler_service = profiler_service
        self._viewer_tool_updated_once = False
        self._create_connections()

    def _create_connections(self):
        self._create_ramanometer_tool_connections()
        self._create_viewer_tool_connections()
        self._create_logger_tool_connections()
        self._create_scanner_tool_connections()
        self._create_timing_tool_connections()
        self._create_timing_pulse_settings_tool_connections()
        self._create_alignment_tool_connections()
        self._create_profiler_tool_connections()
        self._create_profiler_pulse_settings_tool_connections()

    def _create_ramanometer_tool_connections(self):
        self._ramanometer_tool_ui.print_to_e_log_action.triggered.connect(self._e_log_printer.show_dialog)
        self._ramanometer_tool_ui.export_action.triggered.connect(self._export)
        self._ramanometer_tool_ui.tab_widget.currentChanged.connect(self._offline_update_tools)
        self._ramanometer_tool_ui.tab_widget.currentChanged.connect(self._update_scans)

    def _create_viewer_tool_connections(self):
        self._create_tool_connections(caller=self._viewer_tool, callee=self._logger_tool)
        self._viewer_tool.ui.take_button.clicked.connect(self._start_background_taking_process_loop)

    def _create_logger_tool_connections(self):
        self._create_tool_connections(caller=self._logger_tool, callee=self._viewer_tool)

    def _create_tool_connections(self, caller, callee):
        caller.ui.start_button.clicked.connect(self._start_main_process_loop)
        caller.ui.start_button.clicked.connect(callee.start_button_clicked)
        caller.ui.stop_button.clicked.connect(self._stop_main_process_loop)
        caller.ui.stop_button.clicked.connect(callee.stop_button_clicked)
        caller.ui.stop_button.clicked.connect(self._scanner_tool.stop_button_clicked)
        caller.ui.stop_button.clicked.connect(self._alignment_tool.stop_button_clicked)

    def _create_scanner_tool_connections(self):
        self._scanner_tool.ui.start_button.clicked.connect(self._start_scanning_process_loop)
        self._scanner_tool.ui.stop_button.clicked.connect(self._stop_main_process_loop)
        self._scanner_tool.ui.stop_button.clicked.connect(self._viewer_tool.stop_button_clicked)
        self._scanner_tool.ui.stop_button.clicked.connect(self._logger_tool.stop_button_clicked)
        self._scanner_tool.ui.stop_button.clicked.connect(self._alignment_tool.stop_button_clicked)

    def _create_timing_tool_connections(self):
        self._timing_tool.ui.check_button.clicked.connect(self._check_timing_pulse_settings)
        self._timing_tool.ui.start_button.clicked.connect(self._start_timing_process_loop)
        self._timing_tool.ui.stop_button.clicked.connect(self._stop_timing_process_loop)
        self._timing_tool.ui.fast_shutter_check_box.toggled.connect(self._timing_pulse_settings_tool.reset)

    def _create_timing_pulse_settings_tool_connections(self):
        self._timing_pulse_settings_tool.ui.update_button.clicked.connect(self._update_timing_pulse_settings)

    def _create_alignment_tool_connections(self):
        self._alignment_tool.ui.start_button.clicked.connect(self._start_alignment_process_loop)
        self._alignment_tool.ui.stop_button.clicked.connect(self._stop_main_process_loop)
        self._alignment_tool.ui.stop_button.clicked.connect(self._viewer_tool.stop_button_clicked)
        self._alignment_tool.ui.stop_button.clicked.connect(self._logger_tool.stop_button_clicked)
        self._alignment_tool.ui.stop_button.clicked.connect(self._scanner_tool.stop_button_clicked)

    def _create_profiler_tool_connections(self):
        self._profiler_tool.ui.check_button.clicked.connect(self._check_profiler_pulse_settings)
        self._profiler_tool.ui.start_button.clicked.connect(self._start_profiler_process_loop)
        self._profiler_tool.ui.stop_button.clicked.connect(self._stop_profiler_process_loop)

    def _create_profiler_pulse_settings_tool_connections(self):
        self._profiler_pulse_settings_tool.ui.update_button.clicked.connect(self._update_profiler_pulse_settings)

    def _start_main_process_loop(self):
        if not self._main_concurrency_service.is_active():
            self._main_concurrency_service.start()

    def _start_background_taking_process_loop(self):
        if self._viewer_tool.is_started:
            self._start_main_process_loop()
            self._logger_tool.start_button_clicked()

    def _start_scanning_process_loop(self):
        if self._scanner_tool.is_started:
            self._start_main_process_loop()
            self._viewer_tool.start_button_clicked()
            self._logger_tool.start_button_clicked()
            self._viewer_tool.ui.activate_roi(False)
            self._viewer_tool.ui.apply_background_correction_check_box.setEnabled(False)
            self._viewer_tool.switch_off_averaging()

    def _start_alignment_process_loop(self):
        if self._alignment_tool.is_started:
            self._start_main_process_loop()
            self._viewer_tool.start_button_clicked()
            self._logger_tool.start_button_clicked()
            self._viewer_tool.ui.activate_roi(False)
            self._viewer_tool.ui.apply_background_correction_check_box.setEnabled(False)
            self._viewer_tool.switch_off_averaging()

    def _stop_main_process_loop(self):
        if self._main_concurrency_service.is_active():
            self._viewer_tool.ui.activate_roi(True)
            self._viewer_tool.recover_apply_background_correction_enabled_state()
            self._viewer_tool.recover_averaging()
            self._main_concurrency_service.stop()

    def _main_process_loop(self):
        self._update_services()
        self._update_tools()

    def _update_services(self):
        self._io_service_update()
        if self._io_service.is_new_data_available:
            self._fit_service.update()
            self._metrics_service.update()

    def _io_service_update(self):
        try:
            self._io_service.update()
        except (self._io_service.IOException,
                self._io_service.NoCameraImageType,
                self._io_service.image_data.ImageSizeChangedWhileTakingBackground,
                self._io_service.image_data.ImageSizeChangedWhileRunning,
                self._io_service.image_data.ImageSizeChangedWhileAveraging) as exc:
            self._viewer_tool.ui.stop_button.click()
            self._logger_tool.ui.stop_button.click()
            self._scanner_tool.ui.stop_button.click()
            self._alignment_tool.ui.stop_button.click()
            if isinstance(exc, self._io_service.image_data.ImageSizeChangedWhileRunning):
                self._viewer_tool.set_apply_background_correction_check_box_to(False)
            if isinstance(exc, self._io_service.image_data.ImageSizeChangedWhileAveraging):
                self._viewer_tool.ui.averaging_check_box.setChecked(False)
            messages.show_error(str(exc))

    def _update_tools(self):
        if self._io_service.is_new_data_available:
            if self._ramanometer_tool_ui.get_current_ui() == self._viewer_tool.ui:
                self._viewer_tool.update()
                self._viewer_tool_updated_once = True
            elif self._ramanometer_tool_ui.get_current_ui() == self._logger_tool.ui:
                self._logger_tool.update()
            elif self._ramanometer_tool_ui.get_current_ui() == self._scanner_tool.ui:
                self._scanner_tool.update()
            elif self._ramanometer_tool_ui.get_current_ui() == self._alignment_tool.ui:
                self._alignment_tool.update()

            if not self._viewer_tool_updated_once:
                self._viewer_tool.update()
                self._viewer_tool_updated_once = True
                if self._scanner_tool.is_started:
                    self._viewer_tool.ui.activate_roi(False)
                if self._alignment_tool.is_started:
                    self._viewer_tool.ui.activate_roi(False)

            self._viewer_tool.update_background_taking_progress()
            self._scanner_tool.update_scanning_progress()
            self._alignment_tool.update_scanning_progress()

    def _offline_update_tools(self):
        if not self._main_concurrency_service.is_active():
            if self._ramanometer_tool_ui.get_current_ui() == self._viewer_tool.ui:
                self._viewer_tool.offline_update()
            elif self._ramanometer_tool_ui.get_current_ui() == self._logger_tool.ui:
                self._logger_tool.offline_update()

    def _update_scans(self):
        if self._ramanometer_tool_ui.get_current_ui() == self._scanner_tool.ui:
            self._scanner_tool.update_scan()
        elif self._ramanometer_tool_ui.get_current_ui() == self._timing_tool.ui:
            self._timing_tool.update_scan()
        elif self._ramanometer_tool_ui.get_current_ui() == self._alignment_tool.ui:
            self._alignment_tool.update_scan()
        elif self._ramanometer_tool_ui.get_current_ui() == self._profiler_tool.ui:
            self._profiler_tool.update_scan()

    def _check_timing_pulse_settings(self):
        try:
            self._timing_pulse_settings_tool.show_dialog()
        except self._io_service.IOException as exc:
            messages.show_error(str(exc))

    def _update_timing_pulse_settings(self):
        try:
            self._timing_pulse_settings_tool.update()
        except self._io_service.IOException as exc:
            messages.show_error(str(exc))

    def _start_timing_process_loop(self):
        if self._timing_tool.is_started and not self._timing_concurrency_service.is_active():
            try:
                self._timing_pulse_settings_tool.initialize()
                self._timing_service.initialize()
                self._timing_concurrency_service.start()
            except (self._io_service.IOException,
                    self._timing_service.reflectivity_data.UtilizedNumberOfPulsesBelowMinimumAcceptable) as exc:
                self._timing_tool.ui.stop_button.click()
                messages.show_error(str(exc))

    def _stop_timing_process_loop(self):
        if self._timing_concurrency_service.is_active():
            self._timing_concurrency_service.stop()

    def _timing_process_loop(self):
        self._update_timing_service()
        self._update_timing_tool()

    def _update_timing_service(self):
        try:
            self._timing_service.update()
        except self._io_service.IOException as exc:
            self._timing_tool.ui.stop_button.click()
            messages.show_error(str(exc))

    def _update_timing_tool(self):
        if self._timing_service.is_new_data_available:
            if self._ramanometer_tool_ui.get_current_ui() == self._timing_tool.ui:
                self._timing_tool.update()
            self._timing_tool.update_timing_progress()

    def _check_profiler_pulse_settings(self):
        try:
            self._profiler_pulse_settings_tool.show_dialog()
        except self._io_service.IOException as exc:
            messages.show_error(str(exc))

    def _update_profiler_pulse_settings(self):
        try:
            self._profiler_pulse_settings_tool.update()
        except self._io_service.IOException as exc:
            messages.show_error(str(exc))

    def _start_profiler_process_loop(self):
        if self._profiler_tool.is_started and not self._profiler_concurrency_service.is_active():
            try:
                self._profiler_pulse_settings_tool.initialize()
                self._profiler_concurrency_service.start()
            except self._io_service.IOException as exc:
                self._profiler_tool.ui.stop_button.click()
                messages.show_error(str(exc))

    def _stop_profiler_process_loop(self):
        if self._profiler_concurrency_service.is_active():
            self._profiler_concurrency_service.stop()

    def _profiler_process_loop(self):
        self._update_profiler_service()
        self._update_profiler_tool()

    def _update_profiler_service(self):
        try:
            self._profiler_service.update()
        except self._io_service.IOException as exc:
            self._profiler_tool.ui.stop_button.click()
            messages.show_error(str(exc))

    def _update_profiler_tool(self):
        if self._profiler_service.is_new_data_available:
            if self._ramanometer_tool_ui.get_current_ui() == self._profiler_tool.ui:
                self._profiler_tool.update()
            self._profiler_tool.update_profiler_progress()

    def show_window(self):
        self._ramanometer_tool_ui.show_window()

    def _export(self):
        file_name = self._ramanometer_tool_ui.get_current_ui().name
        if self._ramanometer_tool_ui.get_current_ui() == self._logger_tool.ui:
            self._export_service.export(file_name, self._logger_tool.get_export_content())
        elif self._ramanometer_tool_ui.get_current_ui() == self._scanner_tool.ui:
            self._export_service.export(file_name, self._scanner_tool.get_export_content())
        elif self._ramanometer_tool_ui.get_current_ui() == self._timing_tool.ui:
            self._export_service.export(file_name, self._timing_tool.get_export_content())
        elif self._ramanometer_tool_ui.get_current_ui() == self._profiler_tool.ui:
            self._export_service.export(file_name, self._profiler_tool.get_export_content())
        else:
            messages.show_error(f"Not available for '{self._ramanometer_tool_ui.get_current_ui().name}' tool")
