from PyQt5.QtCore import QRegExp, Qt
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import (QDialog, QVBoxLayout, QHBoxLayout, QGroupBox, QPushButton, QLabel, QSizePolicy, QSlider,
                             QLineEdit)

from ..plotter.timing_pulse_settings_plotter import TimingPulseSettingsPlotter


class TimingPulseSettingsToolUi(QDialog):

    def __init__(self):
        super().__init__()
        self._main_layout = QVBoxLayout()
        self._timing_pulse_settings_plotter = TimingPulseSettingsPlotter()
        self._threshold_slider = QSlider(Qt.Horizontal)
        self._update_button = self._get_push_button_with("Update")
        self._laser_pulses_label = QLabel()
        self._fel_pulses_label = QLabel()
        self._utilized_pulses_label = QLabel()
        self._signal_begin_line_edit = self._get_line_edit_with("-1")
        self._signal_end_line_edit = self._get_line_edit_with("1")
        self._background_begin_line_edit = self._get_line_edit_with("-30")
        self._background_end_line_edit = self._get_line_edit_with("-20")
        self._create_layout()

    @staticmethod
    def _get_push_button_with(label):
        push_button = QPushButton(label)
        push_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        push_button.setAutoDefault(False)
        return push_button

    @staticmethod
    def _get_line_edit_with(default_value):
        integer_line_edit = QLineEdit(default_value)
        integer_line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        integer_line_edit.setAlignment(Qt.AlignRight)
        integer_line_edit.setMaxLength(3)
        integer_line_edit.setMaximumWidth(40)
        integer_line_edit.setValidator(QRegExpValidator(QRegExp("[-]?[0-9][0-9]")))
        return integer_line_edit

    def _create_layout(self):
        self._create_main_layout()
        self._create_plot_layout()
        self._create_controls_layout()

    def _create_main_layout(self):
        self.setLayout(self._main_layout)
        self.setModal(True)

    def _create_plot_layout(self):
        plot_layout = QHBoxLayout()
        plot_layout.addWidget(self._timing_pulse_settings_plotter.widget)
        self._main_layout.addLayout(plot_layout)

    def _create_controls_layout(self):
        controls_layout = QHBoxLayout()
        controls_layout.addWidget(self._create_threshold_group_box())
        controls_layout.addWidget(self._create_data_group_box())
        controls_layout.addWidget(self._create_laser_pulses_group_box())
        controls_layout.addWidget(self._create_fel_pulses_group_box())
        controls_layout.addWidget(self._create_utilized_pulses_group_box())
        controls_layout.addWidget(self._create_signal_integration_group_box())
        controls_layout.addWidget(self._create_background_integration_group_box())
        self._main_layout.addLayout(controls_layout)

    def _create_threshold_group_box(self):
        group_box = QGroupBox("Threshold")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._threshold_slider)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_data_group_box(self):
        group_box = QGroupBox("Data")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._update_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_laser_pulses_group_box(self):
        group_box = QGroupBox("Laser")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Pulses:"))
        widgets_layout.addWidget(self._laser_pulses_label)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_fel_pulses_group_box(self):
        group_box = QGroupBox("FEL")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Pulses:"))
        widgets_layout.addWidget(self._fel_pulses_label)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_utilized_pulses_group_box(self):
        group_box = QGroupBox("Utilized")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Pulses:"))
        widgets_layout.addWidget(self._utilized_pulses_label)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_signal_integration_group_box(self):
        group_box = QGroupBox("Signal (relative to peak)")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Begin:"))
        widgets_layout.addWidget(self._signal_begin_line_edit)
        widgets_layout.addWidget(QLabel("End:"))
        widgets_layout.addWidget(self._signal_end_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_background_integration_group_box(self):
        group_box = QGroupBox("Background (relative to peak)")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Begin:"))
        widgets_layout.addWidget(self._background_begin_line_edit)
        widgets_layout.addWidget(QLabel("End:"))
        widgets_layout.addWidget(self._background_end_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    @property
    def threshold_slider(self):
        return self._threshold_slider

    @property
    def update_button(self):
        return self._update_button

    @property
    def signal_begin_line_edit(self):
        return self._signal_begin_line_edit

    @property
    def signal_end_line_edit(self):
        return self._signal_end_line_edit

    @property
    def background_begin_line_edit(self):
        return self._background_begin_line_edit

    @property
    def background_end_line_edit(self):
        return self._background_end_line_edit

    def show_window(self):
        self.setWindowTitle("Timing Pulse Settings")
        self.show()
        self.setMinimumSize(self.width(), self.height())

    def plot_adc(self, adc):
        self._timing_pulse_settings_plotter.set_adc(adc)

    def plot_threshold(self, threshold):
        self._timing_pulse_settings_plotter.set_threshold(threshold)

    def plot_positions(self, positions):
        self._timing_pulse_settings_plotter.set_positions(positions)

    def set_laser_pulses(self, pulses):
        self._laser_pulses_label.setText(str(pulses))

    def set_fel_pulses(self, pulses):
        self._fel_pulses_label.setText(str(pulses))

    def set_utilized_pulses(self, pulses):
        self._utilized_pulses_label.setText(str(pulses))
