from PyQt5.QtCore import Qt, QRegExp
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import (QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QSizePolicy, QProgressBar,
                             QLineEdit, QGroupBox, QLabel, QCheckBox)

from ..plotter.timing_plotter import TimingPlotter


class TimingToolUi(QMainWindow):
    START = "Start"

    def __init__(self):
        super().__init__()
        self._name = "Timing"
        self._main_layout = QVBoxLayout()
        self._timing_plotter = TimingPlotter()
        self._start_button = self._get_push_button_with(self.START)
        self._stop_button = self._get_push_button_with("Stop")
        self._progress_bar = self._get_progress_bar()
        self._begin_line_edit = self._get_line_edit()
        self._end_line_edit = self._get_line_edit()
        self._step_size_line_edit = self._get_line_edit_with("100")
        self._samples_per_step_line_edit = self._get_line_edit_with("10")
        self._fast_shutter_check_box = QCheckBox("Fast Shutter")
        self._check_button = self._get_push_button_with("Check")
        self._create_layout()

    @staticmethod
    def _get_push_button_with(label):
        push_button = QPushButton(label)
        push_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        return push_button

    @staticmethod
    def _get_progress_bar():
        progress_bar = QProgressBar()
        progress_bar.setTextVisible(False)
        return progress_bar

    @staticmethod
    def _get_line_edit():
        line_edit = QLineEdit()
        line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        line_edit.setAlignment(Qt.AlignRight)
        line_edit.setValidator(QRegExpValidator(QRegExp("[-]?[0-9]*[.][0-9]*")))
        line_edit.setMaxLength(6)
        line_edit.setMaximumWidth(60)
        return line_edit

    @staticmethod
    def _get_line_edit_with(default_value):
        integer_line_edit = QLineEdit(default_value)
        integer_line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        integer_line_edit.setAlignment(Qt.AlignRight)
        integer_line_edit.setMaxLength(4)
        integer_line_edit.setMaximumWidth(40)
        integer_line_edit.setValidator(QRegExpValidator(QRegExp("[1-9][0-9]*")))
        return integer_line_edit

    def _create_layout(self):
        self._create_main_layout()
        self._create_plot_layout()
        self._create_controls_layout()

    def _create_main_layout(self):
        main_widget = QWidget()
        main_widget.setLayout(self._main_layout)
        self.setCentralWidget(main_widget)

    def _create_plot_layout(self):
        plot_layout = QHBoxLayout()
        plot_layout.addWidget(self._timing_plotter.widget)
        self._main_layout.addLayout(plot_layout)

    def _create_controls_layout(self):
        controls_layout = QHBoxLayout()
        controls_layout.addWidget(self._create_control_group_box())
        controls_layout.addWidget(self._create_progress_group_box())
        controls_layout.addWidget(self._create_timing_range_group_box())
        controls_layout.addWidget(self._create_scan_setup_group_box())
        controls_layout.addWidget(self._create_pulse_settings_group_box())
        self._main_layout.addLayout(controls_layout)

    def _create_control_group_box(self):
        group_box = QGroupBox("Control")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._start_button)
        widgets_layout.addWidget(self._stop_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_progress_group_box(self):
        group_box = QGroupBox("State")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._progress_bar)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_timing_range_group_box(self):
        group_box = QGroupBox("Motor Delay Range")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Begin (ps):"))
        widgets_layout.addWidget(self._begin_line_edit)
        widgets_layout.addWidget(QLabel("End (ps):"))
        widgets_layout.addWidget(self._end_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_scan_setup_group_box(self):
        group_box = QGroupBox("Timing Setup")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Step Size (fs):"))
        widgets_layout.addWidget(self._step_size_line_edit)
        widgets_layout.addWidget(QLabel("Samples / Step:"))
        widgets_layout.addWidget(self._samples_per_step_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_pulse_settings_group_box(self):
        group_box = QGroupBox("Pulse Settings")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._fast_shutter_check_box)
        widgets_layout.addWidget(self._check_button)
        group_box.setLayout(widgets_layout)
        return group_box

    @property
    def name(self):
        return self._name

    @property
    def start_button(self):
        return self._start_button

    @property
    def stop_button(self):
        return self._stop_button

    @property
    def begin_line_edit(self):
        return self._begin_line_edit

    @property
    def end_line_edit(self):
        return self._end_line_edit

    @property
    def step_size_line_edit(self):
        return self._step_size_line_edit

    @property
    def samples_per_step_line_edit(self):
        return self._samples_per_step_line_edit

    @property
    def fast_shutter_check_box(self):
        return self._fast_shutter_check_box

    @property
    def check_button(self):
        return self._check_button

    def set_progress(self, value):
        if value >= 0:
            self._progress_bar.setLayoutDirection(Qt.LeftToRight)
        else:
            self._progress_bar.setLayoutDirection(Qt.RightToLeft)
        self._progress_bar.setValue(abs(value))

    def plot_timing(self, delay, reflectivity):
        self._timing_plotter.set_timing(delay, reflectivity)

    def clear(self):
        self._timing_plotter.clear()
        self.clear_fit()

    def plot_fit(self, fit):
        self._timing_plotter.set_fit(fit)

    def clear_fit(self):
        self._timing_plotter.clear_fit()
