from PyQt5.QtCore import Qt, QRegExp
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import (QMainWindow, QVBoxLayout, QPushButton, QSizePolicy, QCheckBox, QStackedWidget,
                             QProgressBar, QLineEdit, QWidget, QHBoxLayout, QGroupBox, QLabel)

from ..plotter.viewer_plotter import ViewerPlotter


class ViewerToolUi(QMainWindow):
    START_LABEL = "Start"

    def __init__(self, roi_data):
        super().__init__()
        self._name = "Viewer"
        self._roi_data = roi_data
        self._main_layout = QVBoxLayout()
        self._viewer_plotter = ViewerPlotter()
        self._start_button = self._get_sized_push_button(self.START_LABEL)
        self._stop_button = self._get_sized_push_button("Stop")
        self._lock_aspect_ratio_checkbox = QCheckBox("Lock Aspect Ratio")
        self._averaging_check_box = QCheckBox("Averaging")
        self._take_button_and_progress_bar_stack = QStackedWidget()
        self._take_button = self._get_sized_push_button("Take")
        self._progress_bar = QProgressBar()
        self._samples_line_edit = self._get_sized_line_edit()
        self._apply_background_correction_check_box = QCheckBox("Apply", enabled=False)
        self._enable_roi_check_box = QCheckBox("Enable")
        self._reset_roi_button = self._get_sized_push_button("Reset")
        self._save_roi_button = self._get_sized_push_button("Save")
        self._clear_roi_button = self._get_sized_push_button("Clear")
        self._create_layout()
        self._create_connections()
        self._image_read_once = False
        self._reset_roi_once = False

    @staticmethod
    def _get_sized_push_button(label):
        push_button = QPushButton(label)
        push_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        return push_button

    @staticmethod
    def _get_sized_line_edit():
        samples_line_edit = QLineEdit("10")
        samples_line_edit.setAlignment(Qt.AlignRight)
        samples_line_edit.setMaxLength(4)
        samples_line_edit.setMaximumWidth(40)
        samples_line_edit.setValidator(QRegExpValidator(QRegExp("[1-9][0-9]*")))
        return samples_line_edit

    def _create_layout(self):
        self._create_main_layout()
        self._create_plot_layout()
        self._create_controls_layout()

    def _create_main_layout(self):
        main_widget = QWidget()
        main_widget.setLayout(self._main_layout)
        self.setCentralWidget(main_widget)

    def _create_plot_layout(self):
        plot_layout = QHBoxLayout()
        plot_layout.addWidget(self._viewer_plotter.widget)
        self._main_layout.addLayout(plot_layout)

    def _create_controls_layout(self):
        controls_layout = QHBoxLayout()
        controls_layout.addWidget(self._create_control_group_box())
        controls_layout.addWidget(self._create_background_correction_group_box())
        controls_layout.addWidget(self._create_region_of_interest_group_box())
        self._main_layout.addLayout(controls_layout)

    def _create_control_group_box(self):
        group_box = QGroupBox("Control")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._start_button)
        widgets_layout.addWidget(self._stop_button)
        widgets_layout.addWidget(self._lock_aspect_ratio_checkbox)
        widgets_layout.addWidget(self._averaging_check_box)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_background_correction_group_box(self):
        group_box = QGroupBox("Background Correction")
        group_box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        widgets_layout = QHBoxLayout()
        self._take_button_and_progress_bar_stack.addWidget(self._take_button)
        self._take_button_and_progress_bar_stack.addWidget(self._progress_bar)
        widgets_layout.addWidget(self._take_button_and_progress_bar_stack)
        widgets_layout.addWidget(QLabel("Samples:"))
        widgets_layout.addWidget(self._samples_line_edit)
        widgets_layout.addWidget(self._apply_background_correction_check_box)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_region_of_interest_group_box(self):
        group_box = QGroupBox("Region of Interest")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._enable_roi_check_box)
        self._enable_roi_check_box.setEnabled(False)
        widgets_layout.addWidget(self._reset_roi_button)
        self._reset_roi_button.setEnabled(False)
        widgets_layout.addWidget(self._save_roi_button)
        self._save_roi_button.setEnabled(False)
        widgets_layout.addWidget(self._clear_roi_button)
        self._clear_roi_button.setEnabled(False)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_connections(self):
        self._lock_aspect_ratio_checkbox.stateChanged.connect(self._viewer_plotter.lock_aspect_ratio)
        self._apply_background_correction_check_box.stateChanged.connect(self._viewer_plotter.reset)
        self._enable_roi_check_box.stateChanged.connect(self._reset_roi_button.setEnabled)
        self._enable_roi_check_box.stateChanged.connect(self._save_roi_button.setEnabled)
        self._enable_roi_check_box.stateChanged.connect(self._clear_roi_button.setEnabled)
        self._enable_roi_check_box.stateChanged.connect(self._viewer_plotter.enable_region_of_interest)
        self._enable_roi_check_box.stateChanged.connect(self._conditional_reset_roi)
        self._reset_roi_button.clicked.connect(self._reset_roi)
        self._save_roi_button.clicked.connect(self._save_roi)
        self._clear_roi_button.clicked.connect(self._viewer_plotter.clear_roi)

    @property
    def name(self):
        return self._name

    @property
    def start_button(self):
        return self._start_button

    @property
    def stop_button(self):
        return self._stop_button

    @property
    def averaging_check_box(self):
        return self._averaging_check_box

    @property
    def take_button(self):
        return self._take_button

    @property
    def samples_line_edit(self):
        return self._samples_line_edit

    @property
    def apply_background_correction_check_box(self):
        return self._apply_background_correction_check_box

    def add_to_connections_for_changes(self, action):
        self._viewer_plotter.add_to_connections_for_changes(action)
        self._enable_roi_check_box.stateChanged.connect(action)

    def get_relevant_image_and_offset(self, image):
        if self._enable_roi_check_box.isChecked():
            return self._viewer_plotter.get_relevant_image_and_offset(image)
        return image, (0, 0)

    def _conditional_reset_roi(self):
        if not self._reset_roi_once:
            self._reset_roi_once = True
            self._reset_roi()

    def _reset_roi(self):
        self._viewer_plotter.set_roi_position(float(self._roi_data.position["x"]), float(self._roi_data.position["y"]))
        self._viewer_plotter.set_roi_size(float(self._roi_data.size["width"]), float(self._roi_data.size["height"]))
        self._viewer_plotter.set_roi_rotation(float(self._roi_data.rotation["angle"]))

    def _save_roi(self):
        self._roi_data.position["x"] = float(self._viewer_plotter.roi.pos()[0])
        self._roi_data.position["y"] = float(self._viewer_plotter.roi.pos()[1])
        self._roi_data.size["width"] = float(self._viewer_plotter.roi.size()[0])
        self._roi_data.size["height"] = float(self._viewer_plotter.roi.size()[1])
        self._roi_data.rotation["angle"] = float(self._viewer_plotter.roi.angle())
        self._roi_data.save()

    def show_progress_bar(self):
        self._progress_bar.setValue(0)
        self._take_button_and_progress_bar_stack.setCurrentWidget(self._progress_bar)

    def show_take_button(self):
        self._take_button_and_progress_bar_stack.setCurrentWidget(self._take_button)

    def plot_image(self, image):
        self._viewer_plotter.set_image(image)
        if not self._image_read_once:
            self._image_read_once = True
            self._enable_roi_check_box.setEnabled(True)

    def plot_projections(self, values, units):
        self._viewer_plotter.set_projections(values, units)

    def plot_fits(self, fits):
        self._viewer_plotter.set_fits(fits)

    def show_parameters(self, fit_parameters, average_samples, readback_energy, readback_position):
        self._viewer_plotter.set_experiment_parameters(average_samples, readback_energy, readback_position)
        self._viewer_plotter.set_fit_parameters(fit_parameters)

    def set_progress(self, value):
        self._progress_bar.setValue(value)

    def activate_roi(self, state):
        if self._image_read_once:
            self._enable_roi_check_box.setEnabled(state)
            if self._enable_roi_check_box.isChecked():
                self._reset_roi_button.setEnabled(state)
                self._save_roi_button.setEnabled(state)
                self._clear_roi_button.setEnabled(state)
                self._viewer_plotter.activate_roi(state)

    def update_axis(self):
        self._viewer_plotter.update_axis()
